from .__variable import ThermodynamicVariable, convert_variable, constants
from .__quality_iteration import CODES_TO_BE_ITERATED, QualityIteration
from abc import ABC, abstractmethod
import warnings


GLOBALCounter = 0
ACTIVE_HANDLERS = []
HANDLER_SUBCLASSES = {}
DEFAULT_UNIT_SYSTEM = "SI WITH C"
SI_UNIT_SYSTEM = "MASS BASE SI"


class BaseHandler(ABC):

    T_0 = 0.
    TC = 0.
    T_trip = 0.

    P_0 = 0.
    PC = 0.
    P_trip = 0.

    H_0 = 0.
    S_0 = 0.

    __unit_system = DEFAULT_UNIT_SYSTEM

    def __init__(self, fluids: list, composition: list, unit_system=DEFAULT_UNIT_SYSTEM):

        # If REFPROP is available the program uses it, otherwise COOLPROP is used

        #   DEFAULT MEASURE UNITS:
        #
        #   Temperature:        [°C]
        #   Pressure:           [MPa]
        #   Density:            [kg/m^3]
        #   Enthalpy:           [kJ/kg]
        #   Entropy:            [kJ/(kg*K)]
        #   Speed:              [m/s]
        #   Kinematic vis.:     [cm^2/s]
        #   Viscosity:          [uPa*s]
        #   Thermal cond.:      [mW/(m*K)]
        #   Surface tension:    [mN/m]
        #   Molar Mass:         [kg/kmol]
        #   Heat Capacity:      [kJ/(kg*K)]

        self.fluids = fluids
        self.composition = composition
        self.unit_system = unit_system

    def set_reference_state(self, T_0=20, P_0=101325, T_0unit="C", P_0unit="Pa", old_unit_system=None):

        T_unit = self.return_units("T")
        P_unit = self.return_units("P")

        if old_unit_system is not None:

            T_unit_old = self.return_units("T", unit_system=old_unit_system)
            P_unit_old = self.return_units("P", unit_system=old_unit_system)

            self.T_0, info = convert_variable(self.T_0, "T",  T_unit_old, T_unit)
            self.P_0, info = convert_variable(self.P_0, "p",  P_unit_old, P_unit)

        else:

            self.T_0, info = convert_variable(T_0, "T", T_0unit, T_unit)
            self.P_0, info = convert_variable(P_0, "p", P_0unit, P_unit)

        self.H_0 = self.base_calculate("TP", "H", self.T_0, self.P_0)
        self.S_0 = self.base_calculate("TP", "s", self.T_0, self.P_0)

    def base_calculate(self, str_in: str, str_out: str, a: float, b: float):

        global GLOBALCounter

        if str_in in CODES_TO_BE_ITERATED:

            qi = QualityIteration(self, str_in, str_out, a, b)
            return qi.result

        if str_out == "Q":

            if self.unit_is_mass_based:

                str_out = "QMASS"

            else:

                str_out = "QMOLE"

        GLOBALCounter += 1
        return self.base_evaluate_dll(str_in, str_out, a, b)

    def calculate(self, str_out: str, variable_a: ThermodynamicVariable, variable_b: ThermodynamicVariable, metastb=""):

        global GLOBALCounter
        GLOBALCounter += 1

        return self.evaluate_dll(str_out, variable_a, variable_b, metastb)

    @abstractmethod
    def evaluate_dll(self, str_out: str, variable_a: ThermodynamicVariable, variable_b: ThermodynamicVariable, metastb=""):
        pass

    @abstractmethod
    def base_evaluate_dll(self, str_in: str, str_out: str, a: float, b: float):
        pass

    def get_composition(self, phase, T, P):

        # TODO
        return self.composition

    def return_units(self, property_name, unit_system=None):

        if unit_system is None:

            return_value = constants.get_units(property_name, self.unit_system)

        else:

            return_value = constants.get_units(property_name, unit_system)

        if "Unknown" in return_value:

            return ""

        else:

            return return_value

    @property
    def unit_system(self):

        try:
            return self.__unit_system

        except:
            return None

    @unit_system.setter
    def unit_system(self, unit_system_input):

        old_unit_system = self.unit_system
        self.__unit_system = unit_system_input

        try:

            self.set_unit_system(unit_system_input)

        except:

            self.set_unit_system(DEFAULT_UNIT_SYSTEM)
            warning_message = (

                """
                    {} unit system is not supported,
                    {} has been used instead\n
                    Check in the REFPROP manual for the correct
                    name of the system that you wanted to use
                """

            ).format(self.__unit_system, DEFAULT_UNIT_SYSTEM)

            warnings.warn(warning_message)
            self.__unit_system = DEFAULT_UNIT_SYSTEM

        # Evaluate Critical and Triple Point
        self.TC = self.base_calculate("", "TC", 0, 0)
        self.PC = self.base_calculate("", "PC", 0, 0)
        self.T_trip = self.base_calculate("", "TTRP", 0, 0)
        self.P_trip = self.base_calculate("", "PTRP", 0, 0)

        self.set_reference_state(old_unit_system=old_unit_system)

    @abstractmethod
    def set_unit_system(self, unit_system_input):

        pass

    @property
    def unit_is_mass_based(self):

        return not ("mol" in self.return_units("s"))

    @property
    def T_0_in_K(self):

        T_unit = self.return_units("T")
        TO, info = convert_variable(self.T_0, "T",  T_unit, "K")
        return TO

    @staticmethod
    def print_global_counter(reset=True):

        global GLOBALCounter
        print(GLOBALCounter)

        if reset:
            GLOBALCounter = 0

    def __init_subclass__(cls, **kwargs):

        global HANDLER_SUBCLASSES
        super().__init_subclass__(**kwargs)
        HANDLER_SUBCLASSES.update({cls.__name__: cls})


def find_handler(input_handler_dict):

    global ACTIVE_HANDLERS

    for handler_dict in ACTIVE_HANDLERS:

        if dicts_are_equal(handler_dict, input_handler_dict):
            return handler_dict["instance"]

    return None


def dicts_are_equal(curr_dict, other_dict):

    for key in ["fluids", "composition", "unit_system", "class"]:

        if not (curr_dict[key] == other_dict[key]):
            return False

    return True


def init_handler(chosen_subclass: type(BaseHandler), fluids: list, composition: list, unit_system=DEFAULT_UNIT_SYSTEM):

    global ACTIVE_HANDLERS

    handler_dict = {

        "class": chosen_subclass,
        "fluids": fluids,
        "composition": composition,
        "unit_system": unit_system

    }

    new_subclass = find_handler(handler_dict)

    if new_subclass is None:

        new_subclass = chosen_subclass(fluids=fluids, composition=composition, unit_system=unit_system)
        handler_dict.update({

            "instance": new_subclass

        })

        ACTIVE_HANDLERS.append(handler_dict)

    return new_subclass



from .Tools.units_converter import convert_variable, constants


class ThermodynamicVariable:

    def __init__(self, name: str):

        self.value = None
        self.name = name
        self.refprop_name = constants.get_refprop_name(name)
        self.is_user_defined = False
        self.order = 0

    @property
    def is_empty(self):
        return self.value is None

    def convert(self, rp_handler, to_unit_system):

        value, info = convert_variable(

            self.value, self.refprop_name,
            rp_handler.return_units(self.refprop_name),
            rp_handler.return_units(self.refprop_name, to_unit_system)

        )

        # TODO implement conversion mass / mole based system

        return value

    def set_from_different_us(self, value, rp_handler, from_unit_system):

        self.value, info = convert_variable(

            value, self.refprop_name,
            rp_handler.return_units(self.refprop_name, from_unit_system),
            rp_handler.return_units(self.refprop_name)

        )

    def __gt__(self, other):
        # enables comparison
        # self > other

        return self.order > other.order

    def __lt__(self, other):
        # enables comparison
        # self < other

        return self.order < other.order

    def __le__(self, other):
        return not self.__gt__(other)

    def __ge__(self, other):
        return not self.__lt__(other)



from copy import deepcopy

CODES_TO_BE_ITERATED = [

    "HQ", "SQ", "EQ",
    "QH", "QS", "QE"

]


class QualityIteration:

    def __init__(self, RPHandler, str_in: str, str_out: str, a: float, b: float):

        self.RPHandler = RPHandler
        self.str_in = str_in

        self.__identify_Q(a, b)
        self.__identify_T_range()
        self.__iterate_quality()

        self.result = self.RPHandler.base_calculate("TQ", str_out, self.T_limits[0], self.q_value)

    def __identify_Q(self, a, b):

        self.other_var = self.str_in.strip("Q")

        if self.str_in[0] == "Q":

            self.q_value = a
            self.var_value = b

        else:

            self.q_value = b
            self.var_value = a

    def __identify_T_range(self):

        min_T = self.RPHandler.base_calculate("EOSMIN", "T", 0., 0.)

        try:

            triple_T = self.RPHandler.base_calculate("TRIP", "T", 0., 0.)

        except:

            triple_T = -1000000

        self.min_T = max(min_T, triple_T)
        self.T_limits = [min_T, self.RPHandler.TC]

    def __iterate_quality(self):

        self.value_limits = [

            self.__calculate_var(self.T_limits[0]),
            self.__calculate_var(self.T_limits[1])

        ]

        if not (self.value_limits[0] * self.value_limits[1] < 0):
            raise Exception("Unable to perform {} flash".format(self.str_in))

        counter = 0
        while abs(self.T_limits[0] - self.T_limits[1]) > 10**-3 or counter > 30:
            counter += 1
            self.__quality_iteration_step()

    def __quality_iteration_step(self):

        # Bisection Calculation
        T_bis = (self.T_limits[1] + self.T_limits[0]) / 2
        bis_dict = self.__get_quality_iteration_step_result(T_bis)
        dict_list = [bis_dict]

        # Secant Calculation (if possible)
        T_val = self.T_limits
        res_val = self.value_limits
        T_sec = T_val[0] - res_val[0] / (res_val[1] - res_val[0]) * (T_val[1] - T_val[0])

        if self.min_T < T_sec < self.RPHandler.TC:

            sec_dict = self.__get_quality_iteration_step_result(T_sec)
            dict_list.append(sec_dict)

            # if possible check the interval between T_sec and T_bis
            if bis_dict["var_new"] * sec_dict["var_new"] < 0:
                dict_list.append({

                    "new_limits": {

                        "T": [bis_dict["T_new"], sec_dict["T_new"]],
                        "var": [bis_dict["var_new"], sec_dict["var_new"]]

                    },
                    "dT": abs(bis_dict["T_new"] - sec_dict["T_new"])

                })

        # take the minimum dt interval
        min_dict = bis_dict
        for res_dict in dict_list:

            if min_dict["dT"] > res_dict["dT"]:
                min_dict = res_dict

        self.T_limits = min_dict["new_limits"]["T"]
        self.value_limits = min_dict["new_limits"]["var"]

    def __get_quality_iteration_step_result(self, new_T):

        T_limits_new = deepcopy(self.T_limits)
        value_limits_new = deepcopy(self.value_limits)
        new_var = self.__calculate_var(new_T)

        if new_var * self.value_limits[1] < 0:

            T_limits_new[0] = new_T
            value_limits_new[0] = new_var

        else:

            T_limits_new[1] = new_T
            value_limits_new[1] = new_var

        return {

            "T_new": new_T,
            "var_new": new_var,
            "new_limits": {

                "T": T_limits_new,
                "var": value_limits_new

            },
            "dT": abs(T_limits_new[1] - T_limits_new[0])

        }

    def __calculate_var(self, T_value):

        value = self.RPHandler.base_calculate("TQ", self.other_var, T_value, self.q_value)
        error = value - self.var_value
        return error


from REFPROPConnector.refprop_calculator import (

    init_handler,
    RefPropHandler,
    AbstractThermodynamicPoint,
    ThermodynamicPoint

)
from REFPROPConnector.Handlers.Tools.refprop_plotter import DiagramPlotter, DiagramPlotterOptions
from REFPROPConnector.Support.resources.file_handler import retreive_RP_exec



import warnings

from scipy.integrate import solve_ivp
from abc import ABC, abstractmethod
from sty import fg, bg, ef, rs
import sty


from REFPROPConnector.Handlers import (

    RefPropHandler, CODES_TO_BE_ITERATED, init_handler,
    ThermodynamicVariable, constants, DEFAULT_UNIT_SYSTEM,
    SI_UNIT_SYSTEM

)


class AbstractThermodynamicPoint(ABC):

    @classmethod
    def init_from_fluid(

            cls, fluids: list, composition: list,
            other_variables="all", calculate_on_need="all",
            unit_system=DEFAULT_UNIT_SYSTEM

    ):

        RP = init_handler(

            chosen_subclass=RefPropHandler,
            fluids=fluids, composition=composition,
            unit_system=unit_system

        )
        return cls(RP, other_variables, calculate_on_need)

    def __init__(self, refprop: RefPropHandler, other_variables="all", calculate_on_need="all"):

        self.RPHandler = refprop
        self.inputs = {

            "other_variables": other_variables,
            "calculate_on_need": calculate_on_need

        }

        self.__initialize_state_variables()
        self.__initialize_other_variables(other_variables)
        self.__reset_state_var_order()

        self.calculated_variables = list()
        self.__initialize_calculate_on_need_variables(calculate_on_need)
        self.__metastability = ""
        self.__tmp_si_point = None

    def __initialize_state_variables(self):

        self.state_var_list = list()

        for variable_name in ["P", "T", "h", "s", "Q", "rho"]:

            new_variable = ThermodynamicVariable(variable_name)
            self.state_var_list.append(new_variable)

    def __initialize_other_variables(self, other_variables):

        self.other_variables = list()

        if other_variables is not None:

            if other_variables == "all":

                other_variables = constants.get_all_refprop_names()

            if type(other_variables) == list:

                state_variables_refprop_names = self.state_variables_refprop_names

                for variable_name in other_variables:

                    if not constants.get_refprop_name(variable_name) in state_variables_refprop_names:

                        new_variable = ThermodynamicVariable(variable_name)
                        self.other_variables.append(new_variable)

    def __initialize_calculate_on_need_variables(self, calculate_on_need):

        self.calculate_on_need_variables = list()

        if calculate_on_need is not None:

            if type(calculate_on_need) == list:

                for varible_name in calculate_on_need:

                    found_variable = self.__get_variable_from_name(varible_name)

                    if found_variable is not None:
                        self.calculate_on_need_variables.append(found_variable)

            elif calculate_on_need == "all":

                self.calculate_on_need_variables.extend(self.variables)

            elif calculate_on_need == "other":

                self.calculate_on_need_variables.extend(self.other_variables)

            elif calculate_on_need == "state":

                self.calculate_on_need_variables.extend(self.state_var_list)

    def __update_variables(self):

        not_none_variables = self.not_none_variables
        self.calculated_variables = not_none_variables

        for variable in self.state_var_list:

            if variable in not_none_variables:

                variable.is_user_defined = True

            else:

                if variable not in self.calculate_on_need_variables:
                    self.__calculate_variable(variable)

        for variable in self.other_variables:

            if variable not in self.calculate_on_need_variables:
                self.__calculate_variable(variable)

        self.other_calculation()

    def __calculate_variable(self, variable):

        if self.calculation_ready:

            variable.value = self.__calculate_direct(variable.refprop_name)
            self.calculated_variables.append(variable)

    def __calculate_direct(self, REFPROP_CODE):

        if self.calculation_ready:

            not_none_variables = self.not_none_variables
            input_str = not_none_variables[0].refprop_name + not_none_variables[1].refprop_name

            if input_str not in CODES_TO_BE_ITERATED:

                value = self.RPHandler.calculate(

                    REFPROP_CODE,
                    not_none_variables[0],
                    not_none_variables[1],
                    metastb=self.__metastability

                )

            else:

                P_value = self.RPHandler.calculate(

                    "P",
                    not_none_variables[0],
                    not_none_variables[1],
                    metastb=self.__metastability

                )

                self.set_variable("P", P_value)
                value = self.__calculate_direct(REFPROP_CODE)

            return value

        return None

    @abstractmethod
    def other_calculation(self):
        pass

    def set_variable(self, variable_name: str, variable_value: float, other_unit_system=None):

        input_refprop_name = constants.get_refprop_name(variable_name)

        for variable in self.state_var_list:

            if variable.refprop_name == input_refprop_name:

                if other_unit_system is None:

                    variable.value = variable_value

                else:

                    variable.set_from_different_us(variable_value, self.RPHandler, other_unit_system)

                variable.order = 0

                break

        self.state_var_list.sort()
        self.__reset_state_var_order()

        if self.calculation_ready:

            self.__update_variables()

    def get_variable(self, variable_name: str, other_unit_system=None):

        variable = self.__get_variable_from_name(variable_name)

        if variable is not None:

            if variable not in self.calculated_variables:

                self.__calculate_variable(variable)

            if other_unit_system is None:

                return variable.value

            else:

                return variable.convert(

                    rp_handler=self.RPHandler,
                    to_unit_system=other_unit_system

                )

        if variable_name == "exergy":

            dh = self.get_variable("h") - self.RPHandler.H_0
            ds = self.get_variable("s") - self.RPHandler.S_0

            return dh - self.RPHandler.T_0_in_K * ds

        try:

            return self.__dict__[variable_name]

        except:

            return None

    def get_derivative(self, num_name: str, den_name: str, fix_name: str):

        if self.calculation_ready:

            info = constants.get_derivative_info(num_name, den_name, fix_name)

            if info is not None:

                rp_codes = info.refprop_codes

                if info.is_fraction:

                    a = self.__calculate_direct(rp_codes[0])
                    b = self.__calculate_direct(rp_codes[1])

                    return - a / b

                elif info.is_inverse:

                    return 1 / self.__calculate_direct(rp_codes[0])

                else:

                    return self.__calculate_direct(rp_codes[0])

        return None

    def get_second_derivative(self, num_name: str, den_name: str, other_den_name=None):

        if self.calculation_ready:

            if other_den_name is not None:

                info = constants.get_second_derivative_info(num_name, den_name, other_den_name)

            else:

                info = constants.get_second_derivative_info(num_name, den_name)

            if info is not None:

                rp_codes = info.refprop_codes
                return self.__calculate_direct(rp_codes[0])

        return None

    def evaluate_RP_code(self, rp_code):

        return self.__calculate_direct(rp_code)

    def get_composition(self, phase):

        if self.calculation_ready:

            Q = self.get_variable("Q")

            if 0 < Q < 1:

                T = self.get_variable("T")
                P = self.get_variable("P")
                self.RPHandler.get_composition(phase, T, P)

            else:

                return self.composition

    def get_unit(self, variable_name: str):

        return self.RPHandler.return_units(variable_name)

    def __get_variable_from_name(self, variable_name: str):

        input_refprop_name = constants.get_refprop_name(variable_name.lower())

        for variable in self.variables:

            if variable.refprop_name == input_refprop_name:
                return variable

        return None

    def __reset_state_var_order(self):

        counter = 1
        for variable in self.state_var_list:
            variable.order = counter
            counter += 1

    @property
    def calculation_ready(self):

        counter = 0

        for variable in self.state_var_list:

            if not variable.is_empty:
                counter += 1

        if counter > 2:

            return True

        elif counter == 2:

            __Q = self.__get_variable_from_name("Q")

            if not __Q.is_empty:

                if (__Q.value <= 1) and (__Q.value >= 0):
                    return True

                else:
                    return False

            return True

        else:

            return False

    @property
    def not_none_variables(self) -> list:

        return_list = list()

        for variable in self.state_var_list:

            if not variable.is_empty:

                return_list.append(variable)

                if len(return_list) == 2:
                    break

        return return_list

    @property
    def state_variables_refprop_names(self) -> list:

        return_list = list()
        for variable in self.state_var_list:

            return_list.append(variable.refprop_name)

        return return_list

    @property
    def variables(self):

        variables = list()
        variables.extend(self.state_var_list)
        variables.extend(self.other_variables)

        return variables

    @property
    def composition(self):

        return self.RPHandler.composition

    @property
    def reference_state(self):

        reference = self.duplicate()
        reference.set_variable("T", self.RPHandler.T_0)
        reference.set_variable("P", self.RPHandler.P_0)

        return reference

    @reference_state.setter
    def reference_state(self, state):

        self.RPHandler.set_reference_state(

            T_0 = state.get_variable("T"),
            P_0 = state.get_variable("P"),
            T_0unit = state.get_unit("T"),
            P_0unit = state.get_unit("P")

        )

    @composition.setter
    def composition(self, new_composition: list):

        self.RPHandler.composition = new_composition

    def print_global_counter(self, reset=True):

        self.RPHandler.print_global_counter(reset=reset)

    @abstractmethod
    def duplicate(self):
        """
            used to create a copy of the point for further calculations
            in programming it in subclasses use the input dictionary to recover original inputs

            don't forget to copy the current state defined by the user:

                self.copy_state_to(new_point)

        """
        pass

    @abstractmethod
    def get_alternative_unit_system(self, new_unit_system):

        rp_handler = init_handler(

            chosen_subclass=RefPropHandler,
            fluids=self.RPHandler.fluids,
            composition=self.RPHandler.composition,
            unit_system=new_unit_system

        )

        tp = ThermodynamicPoint(

            self.RPHandler.fluids,
            self.RPHandler.composition,
            rp_handler=rp_handler,
            unit_system=new_unit_system,
            other_variables=self.inputs["other_variables"],
            calculate_on_need=self.inputs["calculate_on_need"]

        )

        self.copy_state_to(tp)
        return tp

    @property
    def metastability(self):


        if self.__metastability == "L":

            return "Liquid"

        if self.__metastability == "V":

            return "Vapour"

        return "Equilibrium"

    @metastability.setter
    def metastability(self, metastability: str):

        metastability = metastability.lower()
        if metastability == "liq" or metastability == "liquid" or metastability == "v" or metastability == ">":

            self.__metastability = "L"

        elif metastability == "vap" or metastability == "vapour" or metastability == "vapor" or metastability == "v" or metastability == "<":

            self.__metastability = "V"

        else:

            self.__metastability = ""

    def get_static_point(self, speed, keep_metastability=False, integrate=True):

        return self.__get_dynamic_variation(

            mult=-1, speed=speed,
            keep_metastability=keep_metastability,
            integrate=integrate

        )

    def get_stagnation_point(self, speed, keep_metastability=False, integrate=True):

        return self.__get_dynamic_variation(

            mult=1, speed=speed,
            keep_metastability=keep_metastability,
            integrate=integrate

        )

    def __get_dynamic_variation(self, mult, speed, keep_metastability=False, integrate=True):

        other_point = self.duplicate()

        if self.__tmp_si_point is None:
            self.__tmp_si_point = self.get_alternative_unit_system(SI_UNIT_SYSTEM)

        self.copy_state_to(self.__tmp_si_point)

        if keep_metastability:
            other_point.metastability = self.__metastability

        h0 = self.get_variable("H", other_unit_system=SI_UNIT_SYSTEM)
        p0 = self.get_variable("P", other_unit_system=SI_UNIT_SYSTEM)
        s0 = self.get_variable("S", other_unit_system=SI_UNIT_SYSTEM)
        dh_dyn = speed ** 2 / 2

        if not integrate:

            dp_dyn = dh_dyn * self.__tmp_si_point.get_variable("rho")

        else:

            try:

                self.__tmp_si_point.set_variable("H", h0 + mult * dh_dyn)
                self.__tmp_si_point.set_variable("S", s0)
                dp_dyn = mult * (self.__tmp_si_point.get_variable("P") - p0)

            except:

                def f(t, y):

                    self.__tmp_si_point.set_variable("H", h0 + mult * t)
                    self.__tmp_si_point.set_variable("P", p0 + mult * y[0])

                    return self.__tmp_si_point.get_variable("rho")

                try:

                    sol = solve_ivp(f, [0, dh_dyn], [0], t_eval=[dh_dyn])
                    dp_dyn = sol.y[0][0]

                except:

                    warnings.warn("It was impossible to reach integral dynamic solution, constant rho solution returned instead")
                    dp_dyn = dh_dyn * self.__tmp_si_point("rho")

        other_point.set_variable("H", h0 + mult * dh_dyn, other_unit_system=SI_UNIT_SYSTEM)
        other_point.set_variable("P", p0 + mult * dp_dyn, other_unit_system=SI_UNIT_SYSTEM)

        return other_point

    def copy_state_to(self, target_point):

        for variable in self.state_var_list[:2]:

            if variable.is_empty:
                break

            else:

                target_point.set_variable(variable.refprop_name, variable.convert(

                    self.RPHandler, target_point.RPHandler.unit_system

                ))

        return target_point

    def list_properties(self):

        string_to_display = """
        
        -----------------------------------------------------------
        -----------------------------------------------------------
            
                        REFPROP_Connector VARIABLES
                
                (if you need more of them to be implemented
                please contact pietro.ungar@unifi.it)
        
        -----------------------------------------------------------
        -----------------------------------------------------------"""

        string_to_display += """
        
            STATE VARIABLES
            (variables that can be set in order to define the 
            thermodynamic state)
            """

        string_to_display += self.__return_variable_list_str(self.state_var_list)

        string_to_display += """
        -----------------------------------------------------------
        
            OTHER VARIABLES
            """

        string_to_display += self.__return_variable_list_str(self.other_variables)

        print(string_to_display)

    def list_unit_systems(self, simplified_display=False, step=4):

        i = 0

        if simplified_display:

            sty.mute(fg, bg, ef, rs)

        else:

            sty.unmute(fg, bg, ef, rs)

        string_to_display = """

                -----------------------------------------------------------
                -----------------------------------------------------------

                              REFPROP_Connector UNIT SYSTEMS

                        (if you need more of them to be implemented
                        please contact pietro.ungar@unifi.it)"""

        while i < len(self.variables):

            string_to_display += """

                -----------------------------------------------------------
                -----------------------------------------------------------

            """

            if i + step < len(self.variables):

                string_to_display += self.__return_variable_unit_str(self.variables[i: i + step])

            else:

                string_to_display += self.__return_variable_unit_str(self.variables[i:])

            i = i + step

        print(string_to_display)

    def __return_variable_list_str(self, variable_list):

        string_to_display = """
            {:<15} {:<25} {:<10}
            """.format("REFPROP NAME:", "ALTERNATIVE NAMES:", "UNIT:")

        for variable in variable_list:

            refprop_name = variable.refprop_name
            other_std_names = constants.get_other_standard_names(refprop_name)
            unit = self.get_unit(refprop_name)

            string_to_display += """
            {:<15} {:<25}  {:<10}
            """.format(refprop_name.upper(), other_std_names[0], unit)

            string_to_display += """
            """.join("{:<15} {:<25}".format("", std_name) for std_name in other_std_names[1:])
            string_to_display += "\n"

        return string_to_display

    def __return_variable_unit_str(self, variable_list):

        name_format = " {:15s}"
        variable_format = " {:20s}"

        name_bold = ef.bold + name_format + ef.rs
        name_back = bg.li_yellow + name_format + bg.rs
        variable_bold = ef.bold + variable_format + ef.rs
        variable_back = bg.li_yellow + variable_format + bg.rs

        string_to_display = "\n" + name_format.format("") + " ".join(

            variable_bold.format(variable.refprop_name.upper()) for variable in variable_list

        ) + "\n\n"

        for unit_system in constants.get_all_unit_systems():

            if unit_system == self.RPHandler.unit_system:

                string_to_display += name_back.format(unit_system)
                string_to_display += "".join(

                    variable_back.format(constants.get_units(variable.refprop_name, unit_system)) for variable in
                    variable_list

                )

            else:

                string_to_display += name_bold.format(unit_system)
                string_to_display += "".join(

                    variable_format.format(constants.get_units(variable.refprop_name, unit_system)) for variable in
                    variable_list

                )

            string_to_display += "\n"

        return string_to_display

    def __eq__(self, other):

        """check if this is equal to another instance"""
        if not (self.calculation_ready and other.calculation_ready):
            return False

        if not (self.RPHandler.fluids == other.RPHandler.fluids):
            return False

        if not (self.RPHandler.composition == other.RPHandler.composition):
            return False

        if not (self.get_variable("P") == other.get_variable("P", other_unit_system=self.RPHandler.unit_system)):
            return False

        if not (self.get_variable("rho") == other.get_variable("rho", other_unit_system=self.RPHandler.unit_system)):
            return False

        return True


class ThermodynamicPoint(AbstractThermodynamicPoint):

    def __init__(self, fluids: list, composition: list, rp_handler=None, other_variables="all", calculate_on_need="all",
                 unit_system="SI WITH C"):

        if rp_handler is None:

            rp_handler = init_handler(

                chosen_subclass=RefPropHandler,
                fluids=fluids, composition=composition,
                unit_system=unit_system

            )

        super().__init__(rp_handler, other_variables=other_variables, calculate_on_need=calculate_on_need)

    def other_calculation(self):
        pass

    @classmethod
    def init_from_fluid(cls, fluids: list, composition: list, other_variables="all", calculate_on_need="all", unit_system="SI WITH C"):

        return ThermodynamicPoint(fluids, composition, other_variables, calculate_on_need, unit_system)

    def duplicate(self):

        tp = ThermodynamicPoint(

            self.RPHandler.fluids,
            self.RPHandler.composition,
            rp_handler=self.RPHandler,
            unit_system=self.RPHandler.unit_system,
            other_variables=self.inputs["other_variables"],
            calculate_on_need=self.inputs["calculate_on_need"]

        )

        self.copy_state_to(tp)
        return tp

    def get_alternative_unit_system(self, new_unit_system):

        rp_handler = init_handler(

            chosen_subclass=RefPropHandler,
            fluids=self.RPHandler.fluids,
            composition=self.RPHandler.composition,
            unit_system=new_unit_system

        )

        tp = ThermodynamicPoint(

            self.RPHandler.fluids,
            self.RPHandler.composition,
            rp_handler=rp_handler,
            unit_system=new_unit_system,
            other_variables=self.inputs["other_variables"],
            calculate_on_need=self.inputs["calculate_on_need"]

        )

        self.copy_state_to(tp)
        return tp
    


from REFPROPConnector.refprop_calculator import AbstractThermodynamicPoint
from itertools import cycle
from tqdm import tqdm
import numpy as np


class DiagramPlotterOptions:

    def __init__(

            self, n_sat_points=500, n_isolines_points=500,
            x_variable="rho", x_var_range=(1, 1000), x_var_log=True,
            y_variable="P", y_var_range=(4, 15), y_var_log=True,
            isoline_ranges=None, plot_saturation=True

    ):

        if isoline_ranges is None:

            isoline_ranges = ({

                "T": (5, 150, 25),
                "H": (200, 450, 25)

            })

        self.n_sat_points = n_sat_points
        self.n_isolines_points = n_isolines_points
        self.plot_saturation = plot_saturation

        self.x_ax = x_variable
        self.x_ax_rng = x_var_range
        self.x_ax_log = x_var_log

        self.y_ax = y_variable
        self.y_ax_rng = y_var_range
        self.y_ax_log = y_var_log

        self.shown_isolines = isoline_ranges.keys()
        self.isoline_ranges = isoline_ranges
        self.__check_isolines()

    def __check_isolines(self):

        isoline_ranges = dict()
        other_yax = self.other_yax_variable

        if other_yax in self.shown_isolines:
            isoline_ranges.update({other_yax: self.isoline_ranges[other_yax]})

        for isoline_name in self.shown_isolines:
            if (isoline_name not in ["T", "P"]) and (not isoline_name == self.x_ax):
                isoline_ranges.update({isoline_name: self.isoline_ranges[isoline_name]})

        self.isoline_ranges = isoline_ranges
        self.shown_isolines = isoline_ranges.keys()

    @property
    def initialized_sat_values(self):

        return {

            "x": get_np_array(0, 0, 2 * self.n_sat_points),
            "y": get_np_array(0, 0, 2 * self.n_sat_points)

        }

    @property
    def initialized_isolines_values(self):

        isolines_dict = dict()

        if self.other_yax_variable in self.shown_isolines:

            isolines_dict.update({self.other_yax_variable: dict()})

        for isoline_name in self.shown_isolines:

            if (isoline_name not in ["T", "P"]) and (not isoline_name == self.x_ax):

                isolines_dict.update({isoline_name: dict()})

        return isolines_dict

    @property
    def other_yax_variable(self):

        if self.y_ax == "P":

            return "T"

        else:

            return "P"

    def get_range(self, n_points, return_x_ax=True):

        if return_x_ax:

            return get_np_array(self.x_ax_rng[0], self.x_ax_rng[1], n_points, log_scale=self.x_ax_log)

        else:

            return get_np_array(self.y_ax_rng[0], self.y_ax_rng[1], n_points, log_scale=self.y_ax_log)

    @property
    def calculate_tp_isoline(self):

        return self.other_yax_variable in self.shown_isolines

    @property
    def std_isolines_name(self):

        name_list = list()

        for isoline_name in self.shown_isolines:

            if (isoline_name not in ["T", "P"]) and (not isoline_name == self.x_ax):

                name_list.append(isoline_name)

        return name_list


class DiagramPlotter:

    def __init__(self, point: AbstractThermodynamicPoint, options: DiagramPlotterOptions = None):

        self.support_point = point.duplicate()

        if options is not None:
            self.options = options

        else:
            self.options = DiagramPlotterOptions()

        self.sat_values = self.options.initialized_sat_values
        self.isolines_values = self.options.initialized_isolines_values

    def calculate(self):

        self.sat_values = self.options.initialized_sat_values
        self.isolines_values = self.options.initialized_isolines_values

        self.__calculate_saturation_condition()
        self.__calculate_iso_lines()

    def __calculate_saturation_condition(self):

        if self.options.plot_saturation:

            y_crit = self.__get_y_critical_point()

            if self.options.y_ax_rng[0] < y_crit < self.options.y_ax_rng[1]:

                y_values = get_np_array(

                    self.options.y_ax_rng[0], y_crit,
                    self.options.n_sat_points, log_scale=self.options.y_ax_log

                )

            elif y_crit > self.options.y_ax_rng[1]:

                y_values = get_np_array(

                    self.options.y_ax_rng[0], self.options.y_ax_rng[1],
                    self.options.n_sat_points, log_scale=self.options.y_ax_log

                )

            else:

                self.options.plot_saturation = False
                return

            self.sat_values = self.options.initialized_sat_values
            pbar_sat = tqdm(desc="Calculating Saturation Points", total=self.options.n_sat_points)

            for i in range(self.options.n_sat_points):

                self.sat_values["y"][i] = y_values[i]
                self.sat_values["y"][-1 - i] = y_values[i]

                self.support_point.set_variable(self.options.y_ax, y_values[i])
                self.support_point.set_variable("Q", 0)
                self.sat_values["x"][i] = self.support_point.get_variable(self.options.x_ax)

                self.support_point.set_variable(self.options.y_ax, y_values[i])
                self.support_point.set_variable("Q", 1)
                self.sat_values["x"][-1 - i] = self.support_point.get_variable(self.options.x_ax)

                pbar_sat.update(1)

            pbar_sat.close()

    def __calculate_iso_lines(self):

        if self.options.calculate_tp_isoline:

            self.__calculate_iso_tp_lines()

        for isoline_name in self.options.std_isolines_name:

            self.__calculate_iso_val_lines(isoline_name)

    def __calculate_iso_tp_lines(self):

        if self.options.calculate_tp_isoline:

            y_max = self.options.y_ax_rng[1] * 2
            tp_var = self.options.other_yax_variable
            x_var_arr = self.options.get_range(self.options.n_isolines_points, return_x_ax=True)
            tp_range_spec = self.options.isoline_ranges[tp_var]
            tp_range = get_np_array(range_min=tp_range_spec[0], range_max=tp_range_spec[1], n_elements=tp_range_spec[2])

            if tp_var == "T":

                tp_crit = self.support_point.RPHandler.TC

            else:

                tp_crit = self.support_point.RPHandler.PC

            pbar_iso_t = tqdm(desc="iso-T lines calculation", total=self.options.n_isolines_points * len(tp_range))

            for tp in tp_range:

                x_var_list = list()
                y_var_list = list()

                y_sat = 0.
                x_sat = [0., 0.]

                if tp < tp_crit:

                    self.support_point.set_variable(tp_var, tp)
                    self.support_point.set_variable("Q", 0)
                    x_sat[0] = self.support_point.get_variable(self.options.x_ax)

                    self.support_point.set_variable(tp_var, tp)
                    self.support_point.set_variable("Q", 1)
                    x_sat[1] = self.support_point.get_variable(self.options.x_ax)

                    y_sat = self.support_point.get_variable(self.options.y_ax)

                self.support_point.set_variable(self.options.y_ax, y_max)
                self.support_point.set_variable(tp_var, tp)

                # x_var_list.append(self.support_point.get_variable(self.options.x_ax))
                # y_var_list.append(y_max)

                for var in x_var_arr:

                    if x_sat[0] < var < x_sat[1]:

                        x_var_list.append(var)
                        y_var_list.append(y_sat)

                    else:

                        self.support_point.set_variable(self.options.x_ax, var)
                        y_res = self.support_point.get_variable(self.options.y_ax)

                        if 0 < y_res <= y_max:

                            x_var_list.append(var)
                            y_var_list.append(y_res)

                    pbar_iso_t.update(1)

                if len(x_var_list) > 0:

                    self.isolines_values[tp_var].update({

                        '{:.0f} {}'.format(tp, self.support_point.get_unit(tp_var)): {

                            "y": y_var_list,
                            "x": x_var_list,

                        }

                    })

            pbar_iso_t.close()

    def __calculate_iso_val_lines(self, val):

        x_var_arr = self.options.get_range(self.options.n_isolines_points, return_x_ax=True)
        val_range_spec = self.options.isoline_ranges[val]
        val_range = get_np_array(

            range_min=val_range_spec[0],
            range_max=val_range_spec[1],
            n_elements=val_range_spec[2]

        )

        pbar_iso_val = tqdm(

            desc="iso-{} lines calculation".format(val),
            total=self.options.n_isolines_points * len(val_range)

        )

        for curr_val in val_range:

            x_var_list = list()
            y_var_list = list()
            self.support_point.set_variable(val, curr_val)

            for curr_x_var in x_var_arr:

                self.support_point.set_variable(self.options.x_ax, curr_x_var)
                y_res = self.support_point.get_variable(self.options.y_ax)

                if y_res > 0:

                    x_var_list.append(curr_x_var)
                    y_var_list.append(y_res)

                pbar_iso_val.update(1)

            if len(x_var_list) > 0:

                self.isolines_values[val].update({

                    '{:.0f} {}'.format(curr_val, self.support_point.get_unit(val)): {

                        "y": y_var_list,
                        "x": x_var_list,

                    }

                })

        pbar_iso_val.close()

    def __get_y_critical_point(self):

        if self.options.y_ax == "P":

            return self.support_point.RPHandler.PC

        else:

            return self.support_point.RPHandler.TC

    def plot(self, ax_1):

        # plot_saturation
        if self.options.plot_saturation:
            ax_1.plot(self.sat_values["x"], self.sat_values["y"], linewidth=2, color="black")

        lines = ["--", "-.", ":"]
        line_cycler = cycle(lines)
        isolines = self.isolines_values

        for line_key in isolines.keys():

            width = 0.4
            color = "black"
            line_style = next(line_cycler)

            for key in isolines[line_key].keys():

                lines.append(ax_1.plot(

                    isolines[line_key][key]["x"], isolines[line_key][key]["y"],
                    linewidth=width, color=color, alpha=0.5,
                    linestyle=line_style, label=key

                )[0])

        if self.options.y_ax_log:
            ax_1.set_yscale("log")

        if self.options.x_ax_log:
            ax_1.set_xscale("log")

        ax_1.set_xlim(left=self.options.x_ax_rng[0], right=self.options.x_ax_rng[1])
        ax_1.set_ylim(bottom=self.options.y_ax_rng[0], top=self.options.y_ax_rng[1])

        ax_1.set_xlabel(

            r'${}\ [{}]$'.format(self.options.x_ax, self.support_point.get_unit(self.options.x_ax)),
            fontsize='large', loc='right',

        )

        ax_1.set_ylabel(

            r'${}\ [{}]$'.format(self.options.y_ax, self.support_point.get_unit(self.options.y_ax)),
            fontsize='large', loc='top',

        )

        return ax_1


def get_np_array(range_min, range_max, n_elements=10, log_scale=False):

    if log_scale:
        return np.power(range_max / range_min, np.array(range(n_elements)) / (n_elements - 1)) * range_min

    else:
        return np.array(range(n_elements)) / (n_elements - 1) * (range_max - range_min) + range_min



import warnings

from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary

from REFPROPConnector.Support.resources.file_handler import get_RP_EXEC
from .__base_handler import BaseHandler, DEFAULT_UNIT_SYSTEM, CODES_TO_BE_ITERATED, QualityIteration


class RefPropHandler(BaseHandler):

    def __init__(self, fluids: list, composition: list, unit_system=DEFAULT_UNIT_SYSTEM):

        # If REFPROP is available the program uses it, otherwise COOLPROP is used

        #   DEFAULT MEASURE UNITS:
        #
        #   Temperature:        [°C]
        #   Pressure:           [MPa]
        #   Density:            [kg/m^3]
        #   Enthalpy:           [kJ/kg]
        #   Entropy:            [kJ/(kg*K)]
        #   Speed:              [m/s]
        #   Kinematic vis.:     [cm^2/s]
        #   Viscosity:          [uPa*s]
        #   Thermal cond.:      [mW/(m*K)]
        #   Surface tension:    [mN/m]
        #   Molar Mass:         [kg/kmol]
        #   Heat Capacity:      [kJ/(kg*K)]

        self.refprop = REFPROPFunctionLibrary(get_RP_EXEC())
        self.refprop.SETPATHdll(get_RP_EXEC())

        super().__init__(fluids, composition, unit_system)

    def base_evaluate_dll(self, str_in: str, str_out: str, a: float, b: float):
        self.refprop.SETFLUIDSdll('*'.join(self.fluids))
        return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c

    def evaluate_dll(self, str_out: str, variable_a, variable_b, metastb=""):

        str_in = variable_a.refprop_name + variable_b.refprop_name + metastb
        a = variable_a.value
        b = variable_b.value

        if str_in in CODES_TO_BE_ITERATED:

            qi = QualityIteration(self, str_in, str_out, a, b)
            return qi.result

        if str_out == "Q":

            if self.unit_is_mass_based:

                str_out = "QMASS"

            else:

                str_out = "QMOLE"

        self.refprop.SETFLUIDSdll('*'.join(self.fluids))
        return self.refprop.REFPROP1dll(str_in, str_out, self.SI, 1, a, b, self.composition).c

    def set_unit_system(self, unit_system_input):

        self.SI = self.refprop.GETENUMdll(0, unit_system_input).iEnum

    @property
    def rp_version(self):

        return self.refprop.RPVersion()



from REFPROPConnector.Support.resources.file_handler import get_refprop_name_xml
from abc import ABC, abstractmethod


class __AbstractTree(ABC):
    """

        Abstract class for a Binary tree object.

        Each node is defined by a search payload and an optional payload

        Search payload will be compared during binary research hence it has to be UNIQUE and to support comparison
        operation (i.e. "<"; ">"; "==")

        Payload can be every kind of object, if needed an initialization can be performed overwriting the init_payload
        function


    """

    def __init__(self):

        self.search_value = None
        self.left_tree = None
        self.right_tree = None

    def append_value(self, search_value, payload):

        if self.is_empty:

            self.search_value = search_value
            self.init_payload(payload)

            self.left_tree = self.init_empty_self()
            self.right_tree = self.init_empty_self()

        else:

            if self.search_value < search_value:

                self.right_tree.append_value(search_value, payload)

            elif self.search_value > search_value:

                self.left_tree.append_value(search_value, payload)

    def find_element(self, search_value):

        if self.is_empty:

            return None

        else:

            if self.search_value == search_value:

                return self

            elif self.search_value < search_value:

                return self.right_tree.find_element(search_value)

            else:

                return self.left_tree.find_element(search_value)

    @property
    def is_empty(self):

        return self.search_value is None

    @abstractmethod
    def init_payload(self, payload):

        """

            Abstract method for the initialization of the payload. It is called by the append_value function during the
            initialization of a new node (hence when the program finds the right empty spot to add the new node)

            It is useful to add new attributes to the tree node if needed

        """

    @classmethod
    @abstractmethod
    def init_empty_self(self):

        """

            This method is called by the append_value function during the initialization of the new left and right
            branches of the tree. It has to be overwritten in sub-classes even if the best approach for this overwriting
            is simply to call the __init__ method of the subclass as in the example below:

                @classmethod
                def init_empty_self(cls):

                    return cls()

            more complex initialization can be performed if needed.

            IMPORTANT NOTES:

                - keep self.search_value = None otherwise the software will not identify the initialized node as empty!
                - any modification to self.left_tree and self.right_tree will not be effective as the both will be
                  overwritten by append_value when called.

        """

        pass


class RefPropNamesTree(__AbstractTree):

    def __init__(self):

        super().__init__()
        self.refprop_name = None
        self.units_dict = None

    def init_payload(self, payload):

        self.refprop_name = payload["RP_Name"]
        self.units_dict = payload["units_dict"]

    @classmethod
    def init_empty_self(cls):

        return cls()

    @classmethod
    def initialize_from_xml(cls):

        refprop_names_tree = cls()
        root = get_refprop_name_xml()

        for element in root.findall("refprop_name"):

            ref_prop_name = element.attrib["name"].upper()

            unit_dict = dict()

            for unit in element.find("units").findall("unit"):

                unit_dict.update({unit.attrib["name"]: unit.attrib["unit"]})

            value_dict = {"RP_Name": ref_prop_name, "units_dict": unit_dict}

            refprop_names_tree.append_value(ref_prop_name.lower(), value_dict)

            for std_name in element.find("std_names").findall("std_name"):

                refprop_names_tree.append_value(std_name.attrib["name"].lower(), value_dict)

        return refprop_names_tree

    def get_refprop_name(self, input_std_name):

        element_found = self.find_element(input_std_name)

        if element_found is not None:
            return element_found.refprop_name

        return None

    def get_units_dict(self, input_std_name):

        element_found = self.find_element(input_std_name)

        if element_found is not None:
            return element_found.units_dict

        return None

    def append_refprop_names(self, initial_list: list) -> list:

        if self.is_empty:

            return initial_list

        else:

            new_list = self.right_tree.append_refprop_names(initial_list)
            new_list = self.left_tree.append_refprop_names(new_list)

            if self.refprop_name not in initial_list:
                new_list.append(self.refprop_name)

            return new_list

    def append_other_standard_names(self, initial_list: list, refprop_name) -> list:

        if self.is_empty:

            return initial_list

        else:

            new_list = self.right_tree.append_other_standard_names(initial_list, refprop_name)
            new_list = self.left_tree.append_other_standard_names(new_list, refprop_name)

            if self.refprop_name == refprop_name:

                if self.search_value not in initial_list and not self.search_value == refprop_name:
                    new_list.append(self.search_value)

            return new_list


class RefPropUnitConverterTree(__AbstractTree):

    def __init__(self):

        super().__init__()
        self.refprop_name = None
        self.conversion_dict = None
        self.conversion_function = None
        self.expensive_property_on_numerator = None

    def init_payload(self, payload):

        self.refprop_name = payload["RP_Name"]
        self.conversion_dict = payload["conversion_dict"]
        self.conversion_function = payload["conversion_funct"]

        if payload["ext_qty_location"] == "num":
            self.expensive_property_on_numerator = True

        elif payload["ext_qty_location"] == "den":
            self.expensive_property_on_numerator = False

    @classmethod
    def init_empty_self(cls):

        return cls()

    @classmethod
    def initialize_from_xml(cls):

        refprop_names_tree = cls()
        root = get_refprop_name_xml(get_converter=True)

        for element in root.findall("unit_prop_name"):

            ref_prop_name = element.attrib["name"].upper()
            function_elem = element.find("func")

            conversion_dict = dict()
            units = element.find("units")

            if units is not None:

                for unit in units.findall("unit"):

                    curr_unit_name = unit.attrib["name"]
                    conversion_dict.update({curr_unit_name: dict()})

                    for factor in unit.findall("factor"):

                        if not factor.attrib["name"] == "quantity":

                            value = float(factor.attrib["value"])

                        else:

                            value = factor.attrib["value"]

                        conversion_dict[curr_unit_name].update({factor.attrib["name"].lower(): value})

            value_dict = {

                "RP_Name": ref_prop_name,
                "conversion_dict": conversion_dict,
                "conversion_funct": function_elem.attrib["name"],
                "ext_qty_location": function_elem.attrib["ext_qty_location"],

            }

            refprop_names_tree.append_value(ref_prop_name.lower(), value_dict)

        return refprop_names_tree

    def get_conversion_information(self, input_std_name):

        element_found = self.find_element(input_std_name)
        return element_found


class RefPropDerivativesTree(__AbstractTree):

    def __init__(self):

        super().__init__()

        self.refprop_codes = None
        self.is_inverse = False
        self.is_fraction = False

    def init_payload(self, payload):

        self.refprop_codes = payload["REFPROP_CODE"]

        self.is_inverse = (payload["INVERSE"] == str(True))
        self.is_fraction = (payload["FRACTION"] == str(True))

    @classmethod
    def init_empty_self(cls):

        return cls()

    @classmethod
    def initialize_from_xml(cls):

        derivatives_tree = cls()
        derivatives_list = cls.__init_derivatives_list()

        for der_dict in derivatives_list:

            derivatives_tree.append_value(der_dict["SEARCH_CODE"].upper(), der_dict)

        return derivatives_tree

    @classmethod
    def __init_derivatives_list(cls):

        """

            REFPROP provides the values of the partial derivatives of some variables

            -   For the 'main_derivatives' properties (currently P, T, and D) it allows every combination of respective
                derivatives (i.e. dp/dT fixed D, dT/dP fixed D and so on).
                The syntax if the REFPROP call in this case is:

                    D{1}D{2}

                    where

                        {1} is the name of the numerator
                        {2} is the name of the denominator

                moreover, for 'main_derivatives' also second order derivatives can be evaluated, the code are defined as follows:

                    D2{1}D{2}2
                     - or -
                    D2{1}D{2}{3}

                where

                        {1} is the name of the numerator
                        {2} is the name of the denominator
                        {3} is the name of the other denominator (if exist)

            -   For the 'other_derivatives' properties it allows only the calculation of the derivative of such property
                with respect to one 'main_derivatives' property. The syntax is:

                    D{1}D{2}_{3}

                    where

                        {1} is the name of the numerator
                        {2} is the name of the denominator
                        {3} is another main 'main_derivatives' property that is kept fixed

                It is not possible to evaluate second derivatives with these properties.

        """

        root = get_refprop_name_xml(get_derivatives_xml=True)

        main_der = root.find("main_derivatives_prop")
        other_der = root.find("other_derivatives_prop")

        main_der_list = list()
        other_der_list = list()
        derivatives_list = list()

        for element in main_der.findall("refprop_name"):
            main_der_list.append(element.attrib["name"].upper())

        for element in other_der.findall("refprop_name"):
            other_der_list.append(element.attrib["name"].upper())

        derivatives_list.extend(cls.__init_main_der_list(main_der_list, other_der_list))
        derivatives_list.extend(cls.__init_other_der_list(main_der_list, other_der_list))

        return derivatives_list

    @classmethod
    def __init_main_der_list(cls, main_der_list, other_der_list):

        derivatives_list = list()

        for num_name in main_der_list:

            for den_name in main_der_list:

                if not (num_name == den_name):

                    derivatives_list.append(cls.__get_second_derivatives_dict(num_name, den_name))

                    for fix_name in main_der_list:

                        if not (num_name == fix_name) and not (den_name == fix_name):

                            derivatives_list.append(cls.__get_derivatives_dict(num_name, den_name, fix_name, is_main=True))
                            derivatives_list.append(cls.__get_second_derivatives_dict(num_name, den_name, fix_name))

                    for fix_name in other_der_list:

                        derivatives_list.append(cls.__get_derivatives_dict(num_name, den_name, fix_name, is_fraction=True))

        return derivatives_list

    @classmethod
    def __init_other_der_list(cls, main_der_list, other_der_list):

        derivatives_list = list()

        for num_name in other_der_list:

            for den_name in main_der_list:

                for fix_name in main_der_list:

                    if not (den_name == fix_name):

                        derivatives_list.append(cls.__get_derivatives_dict(num_name, den_name, fix_name))
                        derivatives_list.append(cls.__get_derivatives_dict(den_name, num_name, fix_name, is_inverse=True))

        return derivatives_list

    @staticmethod
    def __get_derivatives_dict(num_name, den_name, fix_name, is_main=False, is_inverse=False, is_fraction=False):

        search_code = "D{}D{}_{}".format(num_name, den_name, fix_name)

        if is_main:

            rp_code = ["D{}D{}".format(num_name, den_name)]

        elif is_inverse:

            rp_code = ["D{}D{}_{}".format(den_name, num_name, fix_name)]

        elif is_fraction:

            rp_code = [

                "D{}D{}_{}".format(fix_name, den_name, num_name),
                "D{}D{}_{}".format(fix_name, num_name, den_name)

            ]

        else:

            rp_code = [search_code]

        return {

            "SEARCH_CODE": search_code,
            "REFPROP_CODE": rp_code,

            "INVERSE": str(is_inverse),
            "FRACTION": str(is_fraction)

        }

    @staticmethod
    def __get_second_derivatives_dict(num_name, den_name, other_den_name = "2"):

        search_code = "D2{}D{}{}".format(num_name, den_name, other_den_name)

        if other_den_name == "2":

            rp_code = [search_code]

        else:

            if other_den_name == "P":

                # P is always the first name in the derivative code
                other_den_name = den_name
                den_name = "P"

            elif other_den_name == "T" and den_name == "D":

                # T is always in front of D in the derivative code
                other_den_name = "D"
                den_name = "T"

            rp_code = ["D2{}D{}{}".format(num_name, den_name, other_den_name)]

        return {

            "SEARCH_CODE": search_code,
            "REFPROP_CODE": rp_code,

            "INVERSE": str(False),
            "FRACTION": str(False)

        }


import os

FILE_VERSION="0.3.5"
__CURR_DIR = os.path.dirname(__file__)
__RP_NAMES_TXT = """<?xml version="1.0" encoding="UTF-8" ?>
<data>
    <info version="{fileversion}"/>
    <names>

        <refprop_name name = "P">

            <std_names>

                <std_name name="Pressure"/>
                <std_name name="Pres"/>
                <std_name name="Press"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "kPa" />
                <unit name = "MOLE SI"          unit = "MPa" />
                <unit name = "MASS SI"          unit = "MPa" />
                <unit name = "SI WITH C"        unit = "MPa" />

                <unit name = "MOLAR BASE SI"    unit = "Pa"  />
                <unit name = "MASS BASE SI"     unit = "Pa"  />
                <unit name = "ENGLISH"          unit = "psia"/>
                <unit name = "MOLAR ENGLISH"    unit = "psia"/>

                <unit name = "MKS"              unit = "kPa" />
                <unit name = "CGS"              unit = "MPa" />
                <unit name = "MIXED"            unit = "psia"/>
                <unit name = "MEUNITS"          unit = "bar" />

            </units>

        </refprop_name>
        <refprop_name name = "T">

            <std_names>

                <std_name name="Temperature"/>
                <std_name name="Temp"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "K" />
                <unit name = "MOLE SI"          unit = "K" />
                <unit name = "MASS SI"          unit = "K" />
                <unit name = "SI WITH C"        unit = "C" />

                <unit name = "MOLAR BASE SI"    unit = "K" />
                <unit name = "MASS BASE SI"     unit = "K" />
                <unit name = "ENGLISH"          unit = "F" />
                <unit name = "MOLAR ENGLISH"    unit = "F" />

                <unit name = "MKS"              unit = "K" />
                <unit name = "CGS"              unit = "K" />
                <unit name = "MIXED"            unit = "K" />
                <unit name = "MEUNITS"          unit = "C" />

            </units>

        </refprop_name>
        <refprop_name name = "D">

            <std_names>

                <std_name name="rho"/>
                <std_name name="density"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "mol/dm^3" />
                <unit name = "MOLE SI"          unit = "mol/dm^3" />
                <unit name = "MASS SI"          unit = "kg/m^3" />
                <unit name = "SI WITH C"        unit = "kg/m^3" />

                <unit name = "MOLAR BASE SI"    unit = "mol/m^3" />
                <unit name = "MASS BASE SI"     unit = "kg/m^3" />
                <unit name = "ENGLISH"          unit = "lbm/ft^3" />
                <unit name = "MOLAR ENGLISH"    unit = "lbmol/ft^3" />

                <unit name = "MKS"              unit = "kg/m^3" />
                <unit name = "CGS"              unit = "g/cm^3" />
                <unit name = "MIXED"            unit = "g/cm^3" />
                <unit name = "MEUNITS"          unit = "g/cm^3" />

            </units>

        </refprop_name>
        <refprop_name name = "E">

            <std_names>

                <std_name name="u"/>
                <std_name name="internal_energy"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/mol" />
                <unit name = "MOLE SI"          unit = "J/mol" />
                <unit name = "MASS SI"          unit = "J/g" />
                <unit name = "SI WITH C"        unit = "J/g" />

                <unit name = "MOLAR BASE SI"    unit = "J/mol" />
                <unit name = "MASS BASE SI"     unit = "J/kg" />
                <unit name = "ENGLISH"          unit = "BTU/lbm" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/lbmol" />

                <unit name = "MKS"              unit = "J/g" />
                <unit name = "CGS"              unit = "J/g" />
                <unit name = "MIXED"            unit = "J/g" />
                <unit name = "MEUNITS"          unit = "J/g" />

            </units>

        </refprop_name>
        <refprop_name name = "H">

            <std_names>

                <std_name name="enthalpy"/>
                <std_name name="enth"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/mol" />
                <unit name = "MOLE SI"          unit = "J/mol" />
                <unit name = "MASS SI"          unit = "J/g" />
                <unit name = "SI WITH C"        unit = "J/g" />

                <unit name = "MOLAR BASE SI"    unit = "J/mol" />
                <unit name = "MASS BASE SI"     unit = "J/kg" />
                <unit name = "ENGLISH"          unit = "BTU/lbm" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/lbmol" />

                <unit name = "MKS"              unit = "J/g" />
                <unit name = "CGS"              unit = "J/g" />
                <unit name = "MIXED"            unit = "J/g" />
                <unit name = "MEUNITS"          unit = "J/g" />

            </units>

        </refprop_name>
        <refprop_name name = "S">

            <std_names>

                <std_name name="entropy"/>
                <std_name name="entr"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/(mol * K)" />
                <unit name = "MOLE SI"          unit = "J/(mol * K)" />
                <unit name = "MASS SI"          unit = "J/(g * K)" />
                <unit name = "SI WITH C"        unit = "J/(g * K)" />

                <unit name = "MOLAR BASE SI"    unit = "J/(mol * K)" />
                <unit name = "MASS BASE SI"     unit = "J/(kg * K)" />
                <unit name = "ENGLISH"          unit = "BTU/(lbm * R)" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/(lbmol * R)" />

                <unit name = "MKS"              unit = "J/(g * K)" />
                <unit name = "CGS"              unit = "J/(g * K)" />
                <unit name = "MIXED"            unit = "J/(g * K)" />
                <unit name = "MEUNITS"          unit = "J/(g * K)" />

            </units>

        </refprop_name>
        <refprop_name name = "Q">

            <std_names>

                <std_name name="quality"/>
                <std_name name="x"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = " - " />
                <unit name = "MOLE SI"          unit = " - " />
                <unit name = "MASS SI"          unit = " - " />
                <unit name = "SI WITH C"        unit = " - " />

                <unit name = "MOLAR BASE SI"    unit = " - " />
                <unit name = "MASS BASE SI"     unit = " - " />
                <unit name = "ENGLISH"          unit = " - " />
                <unit name = "MOLAR ENGLISH"    unit = " - " />

                <unit name = "MKS"              unit = " - " />
                <unit name = "CGS"              unit = " - " />
                <unit name = "MIXED"            unit = " - " />
                <unit name = "MEUNITS"          unit = " - " />

            </units>

        </refprop_name>
        <refprop_name name = "VIS">

            <std_names>

                <std_name name="mu"/>
                <std_name name="viscosity"/>
                <std_name name="visc"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "uPa * s" />
                <unit name = "MOLE SI"          unit = "uPa * s" />
                <unit name = "MASS SI"          unit = "uPa * s" />
                <unit name = "SI WITH C"        unit = "uPa * s" />

                <unit name = "MOLAR BASE SI"    unit = "Pa * s" />
                <unit name = "MASS BASE SI"     unit = "Pa * s" />
                <unit name = "ENGLISH"          unit = "lbm/(ft * s)" />
                <unit name = "MOLAR ENGLISH"    unit = "lbm/(ft * s)" />

                <unit name = "MKS"              unit = "uPa * s" />
                <unit name = "CGS"              unit = "uPa * s" />
                <unit name = "MIXED"            unit = "uPa * s" />
                <unit name = "MEUNITS"          unit = "cpoise" />

            </units>

        </refprop_name>
        <refprop_name name = "TCX">

            <std_names>

                <std_name name="thermal_conductivity"/>
                <std_name name="k"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "W/(m * K)" />
                <unit name = "MOLE SI"          unit = "mW/(m * K)" />
                <unit name = "MASS SI"          unit = "mW/(m * K)" />
                <unit name = "SI WITH C"        unit = "mW/(m * K)" />

                <unit name = "MOLAR BASE SI"    unit = "W/(m * K)" />
                <unit name = "MASS BASE SI"     unit = "W/(m * K)" />
                <unit name = "ENGLISH"          unit = "BTU/(h * ft * R)" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/(h * ft * R)" />

                <unit name = "MKS"              unit = "W/(m * K)" />
                <unit name = "CGS"              unit = "mW/(m * K)" />
                <unit name = "MIXED"            unit = "mW/(m * K)" />
                <unit name = "MEUNITS"          unit = "mW/(m * K)" />

            </units>

        </refprop_name>
        <refprop_name name = "CP">

            <std_names>

                <std_name name="cp"/>
                <std_name name="isochoric_heat_capacity"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/(mol * K)" />
                <unit name = "MOLE SI"          unit = "J/(mol * K)" />
                <unit name = "MASS SI"          unit = "J/(g * K)" />
                <unit name = "SI WITH C"        unit = "J/(g * K)" />

                <unit name = "MOLAR BASE SI"    unit = "J/(mol * K)" />
                <unit name = "MASS BASE SI"     unit = "J/(kg * K)" />
                <unit name = "ENGLISH"          unit = "BTU/(lbm * R)" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/(lbmol * R)" />

                <unit name = "MKS"              unit = "J/(g * K)" />
                <unit name = "CGS"              unit = "J/(g * K)" />
                <unit name = "MIXED"            unit = "J/(g * K)" />
                <unit name = "MEUNITS"          unit = "J/(g * K)" />

            </units>

        </refprop_name>
        <refprop_name name = "CP0">

            <std_names>

                <std_name name="cp0"/>
                <std_name name="ideal_isochoric_heat_capacity"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/(mol * K)" />
                <unit name = "MOLE SI"          unit = "J/(mol * K)" />
                <unit name = "MASS SI"          unit = "J/(g * K)" />
                <unit name = "SI WITH C"        unit = "J/(g * K)" />

                <unit name = "MOLAR BASE SI"    unit = "J/(mol * K)" />
                <unit name = "MASS BASE SI"     unit = "J/(kg * K)" />
                <unit name = "ENGLISH"          unit = "BTU/(lbm * R)" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/(lbmol * R)" />

                <unit name = "MKS"              unit = "J/(g * K)" />
                <unit name = "CGS"              unit = "J/(g * K)" />
                <unit name = "MIXED"            unit = "J/(g * K)" />
                <unit name = "MEUNITS"          unit = "J/(g * K)" />

            </units>

        </refprop_name>
        <refprop_name name = "CV">

            <std_names>

                <std_name name="cv"/>
                <std_name name="isobaric_heat_capacity"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "J/(mol * K)" />
                <unit name = "MOLE SI"          unit = "J/(mol * K)" />
                <unit name = "MASS SI"          unit = "J/(g * K)" />
                <unit name = "SI WITH C"        unit = "J/(g * K)" />

                <unit name = "MOLAR BASE SI"    unit = "J/(mol * K)" />
                <unit name = "MASS BASE SI"     unit = "J/(kg * K)" />
                <unit name = "ENGLISH"          unit = "BTU/(lbm * R)" />
                <unit name = "MOLAR ENGLISH"    unit = "BTU/(lbmol * R)" />

                <unit name = "MKS"              unit = "J/(g * K)" />
                <unit name = "CGS"              unit = "J/(g * K)" />
                <unit name = "MIXED"            unit = "J/(g * K)" />
                <unit name = "MEUNITS"          unit = "J/(g * K)" />

            </units>

        </refprop_name>
        <refprop_name name = "CP/CV">

            <std_names>

                <std_name name="gamma"/>
                <std_name name="cp/cv"/>
                <std_name name="heat_capacity_ratio"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = " - " />
                <unit name = "MOLE SI"          unit = " - " />
                <unit name = "MASS SI"          unit = " - " />
                <unit name = "SI WITH C"        unit = " - " />

                <unit name = "MOLAR BASE SI"    unit = " - " />
                <unit name = "MASS BASE SI"     unit = " - " />
                <unit name = "ENGLISH"          unit = " - " />
                <unit name = "MOLAR ENGLISH"    unit = " - " />

                <unit name = "MKS"              unit = " - " />
                <unit name = "CGS"              unit = " - " />
                <unit name = "MIXED"            unit = " - " />
                <unit name = "MEUNITS"          unit = " - " />

            </units>

        </refprop_name>
        <refprop_name name = "W">

            <std_names>

                <std_name name="c"/>
                <std_name name="sos"/>
                <std_name name="speed_of_sound"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = "m/s" />
                <unit name = "MOLE SI"          unit = "m/s" />
                <unit name = "MASS SI"          unit = "m/s" />
                <unit name = "SI WITH C"        unit = "m/s" />

                <unit name = "MOLAR BASE SI"    unit = "m/s" />
                <unit name = "MASS BASE SI"     unit = "m/s" />
                <unit name = "ENGLISH"          unit = "ft/s" />
                <unit name = "MOLAR ENGLISH"    unit = "ft/s" />

                <unit name = "MKS"              unit = "m/s" />
                <unit name = "CGS"              unit = "cm/s" />
                <unit name = "MIXED"            unit = "m/s" />
                <unit name = "MEUNITS"          unit = "cm/s" />

            </units>

        </refprop_name>
        <refprop_name name = "PRANDTL">

            <std_names>

                <std_name name="Pr"/>

            </std_names>
            <units>

                <unit name = "DEFAULT"          unit = " - " />
                <unit name = "MOLE SI"          unit = " - " />
                <unit name = "MASS SI"          unit = " - " />
                <unit name = "SI WITH C"        unit = " - " />

                <unit name = "MOLAR BASE SI"    unit = " - " />
                <unit name = "MASS BASE SI"     unit = " - " />
                <unit name = "ENGLISH"          unit = " - " />
                <unit name = "MOLAR ENGLISH"    unit = " - " />

                <unit name = "MKS"              unit = " - " />
                <unit name = "CGS"              unit = " - " />
                <unit name = "MIXED"            unit = " - " />
                <unit name = "MEUNITS"          unit = " - " />

            </units>

        </refprop_name>

    </names>
    <derivatives>

        <main_derivatives_prop>

            <refprop_name name = "P"/>
            <refprop_name name = "T"/>
            <refprop_name name = "D"/>

        </main_derivatives_prop>
        <other_derivatives_prop>

            <refprop_name name = "H"/>
            <refprop_name name = "S"/>

        </other_derivatives_prop>

    </derivatives>
    <unit_conversion>

        <unit_prop_name name = "P">

            <func name="multiply"   ext_qty_location="none"/>

            <units>
                <unit name = "kPa">
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "MPa">
                    <factor name="mult" value="1000"/>
                </unit>
                <unit name = "Pa">
                    <factor name="mult" value="0.001"/>
                </unit>
                <unit name = "bar">
                    <factor name="mult" value="100"/>
                </unit>
                <unit name = "psia">
                    <factor name="mult" value="6.894757293178"/>
                </unit>
            </units>

        </unit_prop_name>
        <unit_prop_name name = "T">

            <func name="sum_mult"   ext_qty_location="none"/>

            <units>
                <unit name = "K">
                    <factor name="sum" value="-273.15"/>
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "C">
                    <factor name="sum" value="0"/>
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "F">
                    <factor name="sum" value="-32"/>
                    <factor name="mult" value="0.5555555555555556"/>
                </unit>
            </units>

        </unit_prop_name>
        <unit_prop_name name = "D">

            <func name="multiply"   ext_qty_location="num"/>

            <units>

                <unit name = "kg/m^3">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" value="mass"/>
                </unit>
                <unit name = "g/cm^3">
                    <factor name="mult" value="1000"/>
                    <factor name="quantity" value="mass"/>
                </unit>
                <unit name = "lbm/ft^3">
                    <factor name="mult" value="16.018"/>
                    <factor name="quantity" value="mass"/>
                </unit>

                <unit name = "mol/dm^3">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" value="mol"/>
                </unit>
                <unit name = "mol/m^3">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" value="mol"/>
                </unit>
                <unit name = "lbmol/ft^3">
                    <factor name="mult" value="16.018"/>
                    <factor name="quantity" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "E">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/g">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/kg">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" value="mass"/>
                </unit>
                <unit name = "BTU/lbm">
                    <factor name="mult" value="2.326"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/mol">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" value="mol"/>
                </unit>
                <unit name = "BTU/lbmol">
                    <factor name="mult" value="2.326"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "H">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/g">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/kg">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" value="mass"/>
                </unit>
                <unit name = "BTU/lbm">
                    <factor name="mult" value="2.326"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/mol">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" value="mol"/>
                </unit>
                <unit name = "BTU/lbmol">
                    <factor name="mult" value="2.326"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "S">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/(g * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/(kg * K)">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "BTU/(lbm * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/(mol * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>
                <unit name = "BTU/(lbmol * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "Q">

            <func name="none"   ext_qty_location="none"/>

        </unit_prop_name>
        <unit_prop_name name = "VIS">

            <func name="multiply"   ext_qty_location="none"/>

            <units>

                <unit name = "uPa * s">
                    <factor name="mult" value="0.000001"/>
                </unit>
                <unit name = "Pa * s">
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "lbm/(ft * s)">
                    <factor name="mult" value="1.4882"/>
                </unit>
                <unit name = "cpoise">
                    <factor name="mult" value="0.001"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "TCX">

            <func name="multiply"   ext_qty_location="none"/>

            <units>

                <unit name = "W/(m * K)">
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "mW/(m * K)">
                    <factor name="mult" value="0.001"/>
                </unit>
                <unit name = "BTU/(h * ft * R)">
                    <factor name="mult" value="1.730735"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "CP">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/(g * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/(kg * K)">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "BTU/(lbm * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/(mol * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>
                <unit name = "BTU/(lbmol * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "CP0">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/(g * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/(kg * K)">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "BTU/(lbm * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/(mol * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>
                <unit name = "BTU/(lbmol * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "CV">

            <func name="multiply"   ext_qty_location="den"/>

            <units>

                <unit name = "J/(g * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "J/(kg * K)">
                    <factor name="mult" value="0.001"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>
                <unit name = "BTU/(lbm * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mass"/>
                </unit>

                <unit name = "J/(mol * K)">
                    <factor name="mult" value="1"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>
                <unit name = "BTU/(lbmol * R)">
                    <factor name="mult" value="4.1868"/>
                    <factor name="quantity" pos="den" value="mol"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "CP/CV">

            <func name="none"   ext_qty_location="none"/>

        </unit_prop_name>
        <unit_prop_name name = "W">

            <func name="multiply"   ext_qty_location="none"/>

            <units>

                <unit name = "m/s">
                    <factor name="mult" value="1"/>
                </unit>
                <unit name = "ft/s">
                    <factor name="mult" value="0.3048"/>
                </unit>
                <unit name = "cm/s">
                    <factor name="mult" value="0.01"/>
                </unit>

            </units>

        </unit_prop_name>
        <unit_prop_name name = "PRANDTL">

            <func name="none"   ext_qty_location="none"/>

        </unit_prop_name>

    </unit_conversion>

</data>

""".format(fileversion=FILE_VERSION)

def generate_rp_name_file():

    rp_names_path = os.path.join(__CURR_DIR, "REFPROP_names.xml")

    with open(rp_names_path, "w") as file:

        file.write(__RP_NAMES_TXT)


# %% IMPORT MODULES
from REFPROPConnector import ThermodynamicPoint
from matplotlib import pyplot as plt
from pandas import DataFrame
import seaborn as sns
import numpy as np
import random
import time


# %% REFERENCE POINT
tp = ThermodynamicPoint(["Carbon Dioxide"], [1])
tp.set_variable("T", 100)
tp.set_variable("P", 0.1)
p = tp.get_variable("P")
t = tp.get_variable("T")
h = tp.get_variable("h")
s = tp.get_variable("s")
rho = tp.get_variable("rho")

n_points = 1000000


# %% T-P Flash
tp_times = np.empty(n_points)
tp_times[:] = np.nan
for i in range(n_points):

    random_mod = random.random() - 0.5
    mult = (1 + random_mod / 10)

    tp.set_variable("T", 100*mult)
    tp.set_variable("P", 0.1*mult)

    start = time.perf_counter()
    tp.get_variable("h")
    tp.get_variable("s")
    tp.get_variable("rho")
    time_tp = (time.perf_counter() - start) * 1000

    if time_tp < 1:
        tp_times[i] = time_tp

sigma = np.nanstd(tp_times)
mean = np.nanmean(tp_times)
tp_times[np.where(np.abs(tp_times - mean) > 6 * sigma)] = np.nan


# %% P-RHO flash
prho_times = np.empty(n_points)
prho_times[:] = np.nan

for i in range(n_points):
    random_mod = random.random() - 0.5
    mult = (1 + random_mod / 10)
    tp.set_variable("P", p * mult)
    tp.set_variable("rho", rho * mult)

    start = time.perf_counter()
    tp.get_variable("h")
    tp.get_variable("T")
    tp.get_variable("s")
    time_prho = (time.perf_counter() - start) * 1000

    if time_prho < 1.5:

        prho_times[i] = time_prho

sigma = np.nanstd(prho_times)
mean = np.nanmean(prho_times)
prho_times[np.where(np.abs(prho_times - mean) > 6 * sigma)] = np.nan


# %% S-H flash
sh_times = np.empty(n_points)
sh_times[:] = np.nan

for i in range(n_points):

    random_mod = random.random() - 0.5
    mult = (1 + random_mod / 10)
    tp.set_variable("S", s * mult)
    tp.set_variable("H", h * mult)

    start = time.perf_counter()
    tp.get_variable("P")
    tp.get_variable("T")
    tp.get_variable("rho")
    time_sh = (time.perf_counter() - start) * 1000

    if time_sh < 1.5:
        sh_times[i] = time_sh

sigma = np.nanstd(sh_times)
mean = np.nanmean(sh_times)
sh_times[np.where(np.abs(sh_times - mean) > 6 * sigma)] = np.nan


# %% P-H flash
ph_times = np.empty(n_points)
ph_times[:] = np.nan

for i in range(n_points):

    random_mod = random.random() - 0.5
    mult = (1 + random_mod / 10)
    tp.set_variable("P", p * mult)
    tp.set_variable("H", h * mult)

    start = time.perf_counter()
    tp.get_variable("h")
    tp.get_variable("T")
    tp.get_variable("rho")
    time_ph = (time.perf_counter() - start) * 1000

    if time_ph < 1.5:
        ph_times[i] = time_ph

sigma = np.nanstd(ph_times)
mean = np.nanmean(ph_times)
ph_times[np.where(np.abs(ph_times - mean) > 6 * sigma)] = np.nan


# %% PLOT_HISTS
data = {

    "T-p flash": tp_times,
    "p-rho flash": prho_times,
    "p-h flash": ph_times,
    "s-h flash": sh_times

}
df = DataFrame(data)
sns.kdeplot(data=df, palette=['orange', 'blue', 'red', 'green'], fill=True)
plt.yscale("log")
plt.show()



from .refprop_names_tree import RefPropUnitConverterTree as __RP_conv_tree
from .refprop_names_tree import RefPropDerivativesTree as __RP_der_tree
from .refprop_names_tree import RefPropNamesTree as __RP_name_tree
import os


CURRENT_DIR = os.path.dirname(__file__)
__REFPROP_NAME_TREE = __RP_name_tree.initialize_from_xml()
__REFPROP_CONV_TREE = __RP_conv_tree.initialize_from_xml()
__REFPROP_DER_TREE = __RP_der_tree.initialize_from_xml()

def get_refprop_name(name: str):

    return __REFPROP_NAME_TREE.get_refprop_name(name.lower())


def get_units(name: str, unit_system: str):

    unit_system = unit_system.upper()
    units_dict = __REFPROP_NAME_TREE.get_units_dict(name.lower())

    if units_dict is not None:

        if unit_system in units_dict.keys():

            return units_dict[unit_system]

        else:

            return "Unknown Unit System {}".format(unit_system)

    else:

        return "Unknown property name {}".format(name)


def get_all_refprop_names():
    return __REFPROP_NAME_TREE.append_refprop_names(list())


def get_all_unit_systems():
    units_dict = __REFPROP_NAME_TREE.get_units_dict("p")

    if units_dict is not None:
        return units_dict.keys()

    return None


def get_other_standard_names(refprop_name):
    return __REFPROP_NAME_TREE.append_other_standard_names(list(), refprop_name)


def get_derivative_info(num_name: str, den_name: str, fix_name: str):

    der_code = "D{}D{}_{}".format(

        get_refprop_name(num_name),
        get_refprop_name(den_name),
        get_refprop_name(fix_name)

    ).upper()

    return __REFPROP_DER_TREE.find_element(der_code)

def get_second_derivative_info(num_name: str, den_name: str, other_den_name="2"):

    if not other_den_name == "2":
        other_den_name = get_refprop_name(other_den_name)

    der_code = "D2{}D{}{}".format(

        get_refprop_name(num_name),
        get_refprop_name(den_name),
        other_den_name

    ).upper()
    return __REFPROP_DER_TREE.find_element(der_code)

def get_conversion_info(variable_name):
    return __REFPROP_CONV_TREE.get_conversion_information(variable_name.lower())




import REFPROPConnector.Support.constants as constants


def convert_variable(variable_value, variable_name, from_unit, to_unit):

    conversion_info = constants.get_conversion_info(variable_name)

    if conversion_info.conversion_function == "multiply":

        result = __conversion_multiply(variable_value, conversion_info, from_unit, to_unit)

    elif conversion_info.conversion_function == "sum_mult":

        result = __conversion_sum_mult(variable_value, conversion_info, from_unit, to_unit)

    elif from_unit == to_unit:

        result = variable_value

    else:

        return None, conversion_info

    return result, conversion_info


def __conversion_multiply(value, conversion_info, from_unit, to_unit):

    multiplier = conversion_info.conversion_dict[from_unit]["mult"]
    tmp_value = value * multiplier

    multiplier = conversion_info.conversion_dict[to_unit]["mult"]
    return tmp_value / multiplier


def __conversion_sum_mult(value, conversion_info, from_unit, to_unit):

    mult = conversion_info.conversion_dict[from_unit]["mult"]
    sum = conversion_info.conversion_dict[from_unit]["sum"]
    tmp_value = (value + sum) * mult

    mult = conversion_info.conversion_dict[to_unit]["mult"]
    sum = conversion_info.conversion_dict[to_unit]["sum"]
    return tmp_value / mult - sum



import warnings
from CoolProp.CoolProp import PropsSI

from . import ThermodynamicVariable
from .__base_handler import BaseHandler, DEFAULT_UNIT_SYSTEM


class CoolpropHandler(BaseHandler):

    def evaluate_dll(self, str_out: str, variable_a: ThermodynamicVariable, variable_b: ThermodynamicVariable, metastb=""):
        pass

    @classmethod
    def coolprop_fluid_name(cls, fluids: list, composition: list):

        if len(fluids) == 1:
            return fluids[0]

        cp_name = "HEOS::{fluid}[{comp}]".format(fluid=fluids[0], comp=composition[0])

        for i in range(len(fluids) - 1):
            cp_name += "&{fluid}[{comp}]".format(fluid=fluids[i + 1], comp=composition[i + 1])

        return cp_name

    def __init__(self, fluids: list, composition: list, unit_system="SI WITH C"):

        self.fluid_name = self.coolprop_fluid_name(fluids, composition)
        super().__init__(fluids, composition, unit_system)

    def base_evaluate_dll(self, str_in: str, str_out: str, a: float, b: float):

        return PropsSI('T','P',101325,'Q',0,'Water')

    def set_unit_system(self, unit_system_input):

        self.__unit_system = unit_system_input

        try:

            self.SI = self.refprop.GETENUMdll(0, self.__unit_system).iEnum

        except:

            self.SI = self.refprop.GETENUMdll(0, DEFAULT_UNIT_SYSTEM).iEnum

            warning_message = (

                """
                    {} unit system is not supported,
                    {} has been used instead\n
                    Check in the REFPROP manual for the correct
                    name of the system that you wanted to use
                """

            ).format(self.__unit_system, DEFAULT_UNIT_SYSTEM)

            warnings.warn(warning_message)
            self.__unit_system = DEFAULT_UNIT_SYSTEM



from REFPROPConnector import ThermodynamicPoint, DiagramPlotter, DiagramPlotterOptions
from REFPROPConnector.Handlers.Tools.units_converter import convert_variable
from REFPROPConnector.Support.constants import get_refprop_name
from matplotlib import pyplot as plt
import numpy as np
import unittest


class TestREFPROPConnector(unittest.TestCase):

    def test_init(self):

        ThermodynamicPoint(["Carbon Dioxide"], [1])
        self.assertEqual(True, True)

    def test_list_properties(self):

        tp = ThermodynamicPoint(["air"], [1])
        tp.list_properties()
        self.assertEqual(True, True)

    def test_list_units(self):

        tp = ThermodynamicPoint(["air"], [1], unit_system="MASS BASE SI")
        tp.list_unit_systems()
        self.assertEqual(True, True)

    def test_search(self):

        rp_name = get_refprop_name("pressure")
        self.assertEqual("P", rp_name)

    def test_convert(self):

        rp_name, info = convert_variable(32, "T", "F", "K")
        self.assertEqual(273.15, rp_name)

        rp_name, info = convert_variable(1, "P", "MPa", "Pa")
        self.assertEqual(1e6, rp_name)

        rp_name, info = convert_variable(1, "H", "J/g", "J/kg")
        self.assertEqual(1e3, rp_name)

    def test_convert_unit_system(self):

        tp = ThermodynamicPoint(["Carbon Dioxide"], [1])
        tp.set_variable("T", 20)
        tp.set_variable("P", 1)

        tp_new = tp.get_alternative_unit_system("MASS BASE SI")

        self.assertEqual(20+273.15, tp_new.get_variable("T"))

    def test_convert_quality(self):

        tp = ThermodynamicPoint(["Carbon Dioxide"], [1])
        tp.set_variable("P", 2)
        tp.set_variable("Q", 1)

        tp_new = tp.get_alternative_unit_system("MASS BASE SI")

        self.assertEqual(tp.get_variable("T")+273.15, tp_new.get_variable("T"))

    def test_copy(self):

        tp = ThermodynamicPoint(["air"], [1])

        tp.set_variable("T", 20)
        tp.set_variable("P", 0.1)

        new_tp = tp.duplicate()
        check_equal = tp.get_variable("h") == new_tp.get_variable("h")

        tp.set_variable("T", 30)
        check_different = not(tp.get_variable("h") == new_tp.get_variable("h"))

        self.assertEqual(True, check_equal and check_different)

    def test_derivative(self):

        tp = ThermodynamicPoint(["carbon dioxide"], [1])

        tp.set_variable("T", 90)
        tp.set_variable("P", 15)

        print(tp.get_derivative("T", "P", "H"))

        self.assertEqual(True, True)

    def test_second_derivative(self):

        tp = ThermodynamicPoint(["carbon dioxide"], [1])

        tp.set_variable("T", 90)
        tp.set_variable("P", 15)

        print(tp.get_second_derivative("T", "P"))

        self.assertEqual(True, True)

    def test_evaluate_direct(self):

        tp = ThermodynamicPoint(["carbon dioxide"], [1])

        tp.set_variable("T", 90)
        tp.set_variable("P", 15)

        a = tp.evaluate_RP_code("D2PDTD")
        b = tp.get_second_derivative("P", "T", "rho")
        print(a)
        print(b)

        self.assertEqual(a, b)

    def test_reference_state(self):

        tp = ThermodynamicPoint(["carbon dioxide"], [1])

        T_ref = 10
        tp_ref = tp.duplicate()
        tp_ref.set_variable("T", T_ref)
        tp_ref.set_variable("P", 0.101325)

        tp.reference_state = tp_ref

        self.assertEqual(tp.reference_state.get_variable("T"), T_ref)

    def test_QH_flash(self):

        tp = ThermodynamicPoint(["water"], [1])

        tp.set_variable("T", 150)
        tp.set_variable("Q", 0)
        h_new = tp.get_variable("h")

        new_state = tp.duplicate()
        new_state.set_variable("Q", 0.05)
        new_state.set_variable("h", h_new)

        print("{} - {}".format(

            tp.get_variable("P"),
            new_state.get_variable("P")

        ))

    def test_PH_flash(self):

        tp = ThermodynamicPoint(["water"], [1])

        tp.set_variable("T", 150)
        tp.set_variable("Q", 0.5)
        h_new = tp.get_variable("h")
        s_new = tp.get_variable("s")

        new_state = tp.duplicate()
        new_state.set_variable("s", s_new)
        new_state.set_variable("h", h_new)

        print("{} - {}".format(

            tp.get_variable("q"),
            new_state.get_variable("q")

        ))

    def test_diagram_plotter(self):

        tp = ThermodynamicPoint(["Carbon Dioxide"], [1])
        options = DiagramPlotterOptions(

            x_variable="T",
            x_var_range = (0, 150), x_var_log=False,
            y_var_range = (4, 15),
            isoline_ranges={

                "rho": (50, 1000, 25),
                "H": (200, 550, 25)

            }

        )
        plotter = DiagramPlotter(tp, options=options)
        plotter.calculate()

        fig, (ax_1) = plt.subplots(1, 1, dpi=200)
        fig.set_size_inches(10, 5)
        plotter.plot(ax_1)
        plt.show()

    def test_equals(self):

        tp = ThermodynamicPoint(["Carbon Dioxide"], [1])
        tp.set_variable("T", 100)
        tp.set_variable("P", 0.1)

        tp_new = tp.get_alternative_unit_system("MASS BASE SI")
        equals = (tp == tp_new)

        self.assertEqual(True, equals)

    def test_metastability(self):

        tp = ThermodynamicPoint(["Carbon Dioxide"], [1])

        t_in = 10
        tp.set_variable("T", t_in)
        tp.set_variable("Q", 0)
        rho_liq = tp.get_variable("rho")

        tp.set_variable("T", t_in)
        tp.set_variable("Q", 1)
        rho_vap = tp.get_variable("rho")

        tp_metastb = tp.duplicate()
        tp_metastb.metastability = "liquid"
        tp_metastb.set_variable("rho", rho_liq * 0.98 + rho_vap * 0.02)
        tp_metastb.set_variable("T", t_in)

        tp.set_variable("rho", rho_liq * 0.98 + rho_vap * 0.02)
        tp.set_variable("T", t_in)
        print(tp.get_variable("P"))
        print(tp_metastb.get_variable("P"))

        self.assertTrue(tp.get_variable("P") > tp_metastb.get_variable("P"))

    def test_dynamic_variation(self):

        tp = ThermodynamicPoint(["Air"], [1])

        t_in = 10
        p_in = 0.101325
        speed = 300

        tp.set_variable("T", t_in)
        tp.set_variable("P", p_in)

        ss = tp.get_variable("W")
        gamma = tp.get_variable("CP/CV")
        ma = speed / ss

        for i in range(500):
            stag_point = tp.get_stagnation_point(speed, integrate=True)

        dp_calc = stag_point.get_variable("P") - p_in
        dp_formula = p_in * (np.power((1 + (gamma - 1) / 2 * ma ** 2), gamma / (gamma - 1)) - 1)

        error_stag = np.abs((dp_calc - dp_formula) / dp_formula)

        static_point = stag_point.get_static_point(speed, integrate=True)
        error_stat = np.abs((static_point.get_variable("P") - p_in) / p_in)

        self.assertTrue((error_stag < 0.001) and (error_stat < 0.001))


if __name__ == '__main__':

    unittest.main()
